#!/usr/bin/env python3
import sys
import os
import docopt
import logging
import termcolor
from . import __doc__ as pkg_doc, __name__ as root_name, DocoptError, __version__
from .parser import ParserParameters, Parser
from .script import Script

log = logging.getLogger(root_name)

__doc__ = pkg_doc + """
Usage:
  docopt.sh generate-library
  docopt.sh [options] [SCRIPT]

Options:
  --prefix PREFIX    Parameter variable name prefix (default: "")
  --line-length N    Max line length when minifying (0 to disable, default: 80)
  --library -l SRC   Generates only the dynamic part of the parser and includes
                     the static parts using `source SRC`, use `generate-library`
                     to create that file
  --no-auto-params   Disable auto-detection parser generation parameters
  --parser           Output parser instead of inserting it in the script
  -h --help          This help message
  --version          Version of this program

Note:
  You can pass the script on stdin as well,
  docopt.sh will then output the modified script to stdout.

Parameters:
  You can set the following global variables before invoking docopt with
  `docopt "$@"` to change the behavior of docopt.

  $DOCOPT_PROGRAM_VERSION  The string to print when --version is specified
                           [default: none/disabled]
  $DOCOPT_ADD_HELP         Set to `false` to not print usage on --help
                           [default: true]
  $DOCOPT_OPTIONS_FIRST    Set to `true` to fail when options are specified
                           after arguments/commands [default: false]
"""


def docopt_sh(params):
  if params['generate-library']:
    parser = Parser(params)
    sys.stdout.write('#!/usr/bin/env bash\n\n' + str(parser.generate_library(check_version=True)))
  else:
    try:
      if params['SCRIPT'] is None:
        if sys.stdin.isatty():
          raise docopt.DocoptExit(
            'Not reading from stdin when it is a tty. '
            'Use either `docopt.sh script.sh` or `docopt.sh < script.sh`.'
          )
        script = Script(sys.stdin.read())
      else:
        with open(params['SCRIPT'], 'r') as h:
          script = Script(h.read(), params['SCRIPT'])
      script.validate()

      parser_parameters = ParserParameters(params, script)
      parser = Parser(parser_parameters)

      if params['--parser']:
        sys.stdout.write(parser.generate(script))
      else:
        patched_script = script.patch(parser)
        if params['SCRIPT'] is None:
          sys.stdout.write(str(patched_script))
        else:
          with open(params['SCRIPT'], 'w') as h:
            h.write(str(patched_script))
          if patched_script == script:
            log.info('The parser in %s is already up-to-date.', params['SCRIPT'])
          else:
            log.info('%s has been updated.', params['SCRIPT'])
    except DocoptError as e:
      log.error(str(e))
      sys.exit(e.exit_code)


def setup_logging():
  level_colors = {
    logging.ERROR: 'red',
    logging.WARN: 'yellow',
  }

  class ColorFormatter(logging.Formatter):

    def format(self, record):
        record.msg = termcolor.colored(record.msg, level_colors.get(record.levelno, None))
        return super(ColorFormatter, self).format(record)

  stderr = logging.StreamHandler(sys.stderr)
  if os.isatty(2):
    stderr.setFormatter(ColorFormatter())
  log.setLevel(level=logging.INFO)
  log.addHandler(stderr)


def main():
  setup_logging()
  params = docopt.docopt(__doc__, version=__version__)
  docopt_sh(params)

if __name__ == '__main__':
  main()



import subprocess
import shlex
from tempfile import NamedTemporaryFile
import os
import io
from contextlib import contextmanager
from docopt_sh.__main__ import main as docopt_sh_main
from docopt_sh.bash import bash_variable_value


def bash_eval_script(script, argv, bash=None):
  argv = ' '.join(map(shlex.quote, argv))
  executable = 'bash' if bash is None else bash[1]
  process = subprocess.run(
    [executable, '-c', 'set - %s; eval "$(cat)"' % argv],
    input=script.encode('utf-8'),
    stdout=subprocess.PIPE, stderr=subprocess.PIPE,
    timeout=2
  )
  return process.returncode, process.stdout.decode('utf-8'), process.stderr.decode('utf-8')


def bash_decl(name, value, bash_version=None):
  if value is None or type(value) in (bool, int, str):
    return 'declare -- {name}={value}'.format(name=name, value=bash_decl_value(value, bash_version))
  if type(value) is list:
    return 'declare -a {name}={value}'.format(name=name, value=bash_decl_value(value, bash_version))
  raise Exception('Unknown value type %s' % type(value))


def bash_decl_value(value, bash_version=None):
  if value is None:
    return '""'
  if type(value) is bool:
    return '"true"' if value else '"false"'
  if type(value) is int:
    return '"{value}"'.format(value=value)
  if type(value) is str:
    return '"{value}"'.format(value=shlex.quote(value).strip("'"))
  if type(value) is list:
    list_tpl = '({value})'
    if bash_version is not None and (int(bash_version[0]) < 4 or int(bash_version[2]) < 4):
      list_tpl = "'({value})'"
    return list_tpl.format(value=' '.join('[{i}]={value}'.format(
      i=i, value=bash_decl_value(v, bash_version)) for i, v in enumerate(value))
    )


def declare_quote(value):
  return value.replace('\\', '\\\\').replace('"', '\\"')


def replace_docopt_params(stream, docopt_params):
  script = stream.read()
  params = '\n'.join([k + '=' + bash_variable_value(v) for k, v in docopt_params.items()])
  script = script.replace('"DOCOPT PARAMS"', params)
  return io.StringIO(script)


@contextmanager
def patched_script(monkeypatch, capsys, name, program_params=[], docopt_params={}, bash=None):
  with open(os.path.join('tests/scripts', name)) as handle:
    script = replace_docopt_params(handle, docopt_params)

  def run(*argv):
    captured = invoke_docopt(monkeypatch, capsys, stdin=script, program_params=program_params)
    return bash_eval_script(captured.out, argv, bash=bash)
  yield run


@contextmanager
def temp_script(name, docopt_params={}, bash=None):
  with open(os.path.join('tests/scripts', name)) as handle:
    script = replace_docopt_params(handle, docopt_params).read()
  file = NamedTemporaryFile(mode='w', delete=False)
  try:
    file.write(script)
    file.close()
    executable = 'bash' if bash is None else bash[1]

    def run(*args):
      process = subprocess.run(
        [executable, file.name] + list(args),
        stdout=subprocess.PIPE, stderr=subprocess.PIPE
      )
      return process.returncode, process.stdout.decode('utf-8'), process.stderr.decode('utf-8')
    yield file, run
  finally:
    os.unlink(file.name)


@contextmanager
def generated_library(monkeypatch, capsys, program_params=[]):
  file = NamedTemporaryFile(mode='w', delete=False)
  try:
    captured = invoke_docopt(monkeypatch, capsys, program_params=['generate-library'] + program_params)
    file.write(captured.out)
    file.close()
    yield file
  finally:
    os.unlink(file.name)


def invoke_docopt(monkeypatch, capsys=None, program_params=[], stdin=None):
  with monkeypatch.context() as m:
    if stdin is not None:
      m.setattr('sys.stdin', stdin)
    m.setattr('sys.argv', ['docopt.sh'] + program_params)
    docopt_sh_main()
    if capsys is not None:
      return capsys.readouterr()
    return None



import logging
import os
import urllib.request
import tarfile
import subprocess
import re
import glob
from .usecases import DocoptUsecaseTestFile

log = logging.getLogger(__name__)


def pytest_addoption(parser):
  parser.addoption(
    '--bash-version', type=str,
    help='test with specific bash version (comma separated, or "all" for all installed versions)'
  )


def pytest_sessionstart(session):
  bash_version = session.config.getoption('--bash-version')
  if bash_version == 'all':
    session.config.bash_versions = get_installed_bash_versions()
    all_versions = [version for version, _ in session.config.bash_versions]
    log.warning('Testing with bash versions: %s' % ', '.join(all_versions))
  elif bash_version:
    session.config.bash_versions = [get_bash_version(version) for version in bash_version.split(',')]
    log.warning('Testing with bash versions: %s' % ', '.join(bash_version.split(',')))
  else:
    version, executable = get_system_bash()
    log.warning('Testing with system bash: %s' % version)
    session.config.bash_versions = [(version, executable)]


def pytest_collect_file(path, parent):
  if path.basename == 'usecases.txt':
    return DocoptUsecaseTestFile(path, parent)


def pytest_generate_tests(metafunc):
  if 'bash' in metafunc.fixturenames:
    if metafunc.config.bash_versions:
      metafunc.parametrize('bash', metafunc.config.bash_versions)


def get_system_bash():
  process = subprocess.run(
    ['bash', '--version'],
    stdout=subprocess.PIPE, stderr=subprocess.PIPE
  )
  process.check_returncode()
  out = process.stdout.decode('utf-8')
  match = re.search(r'GNU bash, version ([^ ]+)', out)
  if not match:
    log.warning('Unable to detect installed bash version from:\n%s\nAssuming 4.4' % out)
    return '4.4', 'bash'
  else:
    return match.group(1), 'bash'


def get_installed_bash_versions():
  versions = [get_system_bash()]
  for executable in glob.glob('tests/bash-versions/bash-*/bash'):
    folder_name = os.path.basename(os.path.dirname(executable))
    versions.append((re.search(r'bash-(.+)', folder_name).group(1), executable))
  return versions


def get_bash_version(version):
  versions_path = 'tests/bash-versions'
  if not os.path.exists(versions_path):
    os.mkdir(versions_path)
  version_path = os.path.join(versions_path, 'bash-%s' % version)
  executable_path = os.path.join(version_path, 'bash')
  if not os.path.exists(executable_path):
    log.warning('bash-%s executable not found in tests/bash-versions, downloading & compiling now' % version)
    archive_path = os.path.join(versions_path, 'bash-%s.tar.gz' % version)
    if os.path.exists(archive_path):
      log.warning('%s already exists, skipping download' % archive_path)
    else:
      url = 'http://ftp.gnu.org/gnu/bash/bash-%s.tar.gz' % version
      urllib.request.urlretrieve(url, archive_path)
    if os.path.exists(version_path):
      log.warning('%s already exists, skipping extraction' % version_path)
    else:
      with tarfile.open(archive_path) as archive:
        archive.extractall(versions_path)
    makefile_path = os.path.join(version_path, 'Makefile')
    if os.path.exists(makefile_path):
      log.warning('%s already exists, skipping ./configure' % makefile_path)
    else:
      process = subprocess.run('./configure', cwd=version_path)
      process.check_returncode()
    process = subprocess.run('make', cwd=version_path)
    process.check_returncode()
  return version, executable_path



import re
import json
import pytest
import subprocess
import shlex
from docopt_sh.script import Script
from docopt_sh.parser import ParserParameters, Parser
from docopt_sh.bash import bash_variable_name
from . import bash_eval_script, bash_decl

import logging
log = logging.getLogger(__name__)


class DocoptUsecaseTestFile(pytest.File):

  def collect(self):
    raw = self.fspath.open().read()
    index = 1
    params = ParserParameters({
      '--prefix': '_',
      '--line-length': None,
      '--library': None,
      '--no-auto-params': False,
    })
    program_template = '''
DOC="{doc}"
DOCOPT_TEARDOWN=false
docopt "$@"
for var in "${{docopt_param_names[@]}}"; do declare -p "$var"; done
'''
    for name, doc, cases in self._parse_test(raw):
      name = self.fspath.purebasename
      if cases:
        script = Script(program_template.format(doc=doc))
        script = str(script.patch(Parser(params)))
      for bash in self.config.bash_versions:
        for case in cases:
          yield DocoptUsecaseTest("%s (%d, bash %s)" % (name, index, bash[0]), self, bash, doc, script, case)
          index += 1

  def _parse_test(self, raw):
    raw = re.compile('#.*$', re.M).sub('', raw).strip()
    if raw.startswith('"""'):
      raw = raw[3:]

    for fixture in raw.split('r"""'):
      name = ''
      doc, _, body = fixture.partition('"""')
      cases = []
      for case in body.split('$')[1:]:
        argv, _, expect = case.strip().partition('\n')
        expect = json.loads(expect)
        prog, _, argv = argv.strip().partition(' ')
        cases.append((prog, argv, expect))

      yield name, doc, cases


class DocoptUsecaseTest(pytest.Item):

  def __init__(self, name, parent, bash, doc, script, case):
    super(DocoptUsecaseTest, self).__init__(name, parent)
    self.doc = doc
    self.script = script
    self.prog, self.argv, self.expect = case
    self.bash = bash
    if type(self.expect) is dict:
      self.expect = {
        bash_variable_name(k, prefix='_'): bash_decl(
          bash_variable_name(k, prefix='_'), v, bash_version=self.bash[0]
        )
        for k, v in self.expect.items()
      }

  def runtest(self):
    try:
      code, out, err = bash_eval_script(self.script, shlex.split(self.argv), bash=self.bash)
      if code == 0:
        expr = re.compile('^declare (--|-a) ([^=]+)=')
        out = out.strip('\n')
        result = {}
        if out != '':
          result = {expr.match(line).group(2): line for line in out.split('\n')}
      else:
        result = 'user-error'
    except Exception as e:
      log.exception(e)

    if self.expect != result:
      if len(err) > 0:
        log.error(self.doc)
        log.error(err)
      raise DocoptUsecaseTestException(self, result)

  def repr_failure(self, excinfo):
    """Called when self.runtest() raises an exception."""
    if isinstance(excinfo.value, DocoptUsecaseTestException):
      return "\n".join((
        "usecase execution failed:",
        self.doc.rstrip(),
        "$ %s %s" % (self.prog, self.argv),
        "result> %s" % json.dumps(excinfo.value.args[1]),
        "expect> %s" % json.dumps(self.expect),
      ))

  def reportinfo(self):
    return self.fspath, 0, "usecase: %s" % self.name


class DocoptUsecaseTestException(Exception):
  pass



from .doc_ast import Option, Command, Required, Optional, OptionsShortcut, OneOrMore, Either
from .bash import Function, bash_variable_name, bash_ifs_value

helper_map = {
  Required: 'docopt_required',
  Optional: 'docopt_optional',
  OptionsShortcut: 'docopt_optional',
  OneOrMore: 'docopt_oneormore',
  Either: 'docopt_either',
}


class Node(Function):

  def __init__(self, pattern, idx):
    self.type = type(pattern)
    self.idx = idx
    super(Node, self).__init__('docopt_node_' + str(idx))


class BranchNode(Node):

  def __init__(self, pattern, idx, function_map):
    super(BranchNode, self).__init__(pattern, idx)
    self.helper_name = helper_map[self.type]
    self.child_indexes = map(lambda child: function_map[child].idx, pattern.children)

  @property
  def body(self):
    # minify arg list by only specifying node idx
    body = ' '.join([self.helper_name] + list(map(str, self.child_indexes)))
    return '  ' + body


class LeafNode(Node):

  def __init__(self, pattern, idx, name_prefix):
    super(LeafNode, self).__init__(pattern, idx)
    self.default_value = pattern.value
    self.pattern = pattern
    if self.type is Option:
      self.helper_name = 'docopt_switch' if type(self.default_value) in [bool, int] else 'docopt_value'
      self.needle = idx
    elif self.type is Command:
      self.helper_name = 'docopt_command'
      self.needle = pattern.name
    else:
      self.helper_name = 'docopt_value'
      self.needle = 'a'
    self.multiple = type(self.default_value) in [list, int]
    self.variable_name = bash_variable_name(pattern.name, name_prefix)

  @property
  def body(self):
    args = [self.variable_name, bash_ifs_value(self.needle)]
    if self.multiple:
      args.append(bash_ifs_value(self.multiple))
    if self.helper_name == 'docopt_command' and args[0] == args[1] and len(args) == 2:
      args = [args[0]]
    body = ' '.join([self.helper_name] + args)
    return '  ' + body



import re
from shlex import quote
from collections import OrderedDict


class Code(object):

  def __init__(self, code):
    self._code = self._get_list(code)

  @property
  def code(self):
    return self._code

  def minify(self, max_line_length):
    return minify(str(self), max_line_length)

  def _get_list(self, elem):
    if type(elem) is list:
      return elem
    elif type(elem) is OrderedDict:
      return list(elem.values())
    elif type(elem) is str:
      return [elem]
    elif isinstance(elem, Function):
      return [elem]
    elif type(elem) is Code:
      return elem.code
    else:
      raise Exception('Unhandled data-type: %s' % type(elem))

  def __add__(self, other):
    return Code(self.code + self._get_list(other))

  def __str__(self):
    return '\n'.join(map(str, self.code))


class Function(Code):

  def __init__(self, name):
    self.name = name

  @property
  def code(self):
    return [str(self)]

  def __str__(self):
    return '{name}(){{\n{body}\n}}\n'.format(name=self.name, body=self.body)

  def __repr__(self):
    lines = self.body.split('\n')
    if len(lines) > 5:
      shortened_body = '\n'.join(lines[:2]) + '\n  ...\n' + '\n'.join(lines[-2:])
    else:
      shortened_body = self.body
    return '{name}(){{\n{body}\n}}'.format(name=self.name, body=shortened_body)


class HelperTemplate(Function):

  def __init__(self, name, function_body):
    self.function_body = function_body
    super(HelperTemplate, self).__init__(name)

  def render(self, replacements={}):
    return Helper(self, replacements)

  @property
  def body(self):
    return self.function_body


class Helper(Function):

  def __init__(self, template, replacements):
    self.template = template
    self.replacements = replacements
    super(Helper, self).__init__(template.name)

  @property
  def body(self):
    body = self.template.function_body
    for placeholder, replacement in self.replacements.items():
      body = body.replace(placeholder, replacement)
    return body


def indent(script, level=1):
  indentation = '  ' * level
  return '\n'.join(map(lambda l: indentation + l, script.split('\n')))


def bash_variable_name(name, prefix=''):
  return prefix + re.sub(r'^[^a-z_]|[^a-z0-9_]', '_', name, 0, re.IGNORECASE)


def bash_variable_value(value):
  if value is None:
    return ''
  if type(value) is bool:
    return 'true' if value else 'false'
  if type(value) is int:
    return str(value)
  if type(value) is str:
    return quote(value)
  if type(value) is list:
    return '(%s)' % ' '.join(bash_variable_value(v) for v in value)
  raise Exception('Unhandled value type %s' % type(value))


def bash_ifs_value(value):
  if value is None or value == '':
    return "''"
  if type(value) is bool:
    return 'true' if value else 'false'
  if type(value) is int:
    return str(value)
  if type(value) is str:
    return quote(value)
  if type(value) is list:
    raise Exception('Unable to convert list to bash value intended for an IFS separated field')
  raise Exception('Unhandled value type %s' % type(value))


def minify(parser_str, max_length):
  lines = parser_str.split('\n')
  lines = remove_leading_spaces(lines)
  lines = remove_empty_lines(lines)
  lines = remove_newlines(lines, max_length)
  return '\n'.join(lines) + '\n'


def remove_leading_spaces(lines):
  for line in lines:
    yield re.sub(r'^\s*', '', line)


def remove_empty_lines(lines):
  for line in lines:
    if line != '':
      yield line


def remove_newlines(lines, max_length):
  def is_comment(line):
    return re.match(r'^\s*#', line) is not None

  def needs_separator(line):
    return re.search(r'; (then|do)$|else$|\{$', line) is None

  def has_continuation(line):
    return re.search(r'\\\s*$', line) is not None

  def remove_continuation(line):
    return re.sub(r'\s*\\\s*$', '', line)

  def combine(line1, line2):
    if is_comment(line1) or is_comment(line2):
      if not is_comment(line1):
        return line1
      if not is_comment(line2):
        return line2
      return None
    if has_continuation(line1):
      return remove_continuation(line1) + ' ' + line2
    if needs_separator(line1):
      return line1 + '; ' + line2
    else:
      return line1 + ' ' + line2

  previous = next(lines)
  for line in lines:
    combined = combine(previous, line)
    if combined is None:
      previous = next(lines, None)
    elif len(combined) > max_length:
      yield previous
      previous = line
    else:
      previous = combined
  if previous:
    yield previous
